from bs4 import BeautifulSoup
import json, logging, glob, os
from collections import namedtuple

my_logger = logging.getLogger(__name__)


class ams_register_calculator:
    """
    This library is made for make easier the conversion between fitbields and register values
    """
    register_struct = namedtuple("register_struct", "address value")


    @classmethod
    def convertRegisterFile(self, path: str) -> dict:
        """
        This method converts a html register file into json and returns the structure as dictionary.
        The converted file is on the same path and with the same name as the source

        Args:
            path (str): path to the file

        Returns:
            dict: dictionary structure of the register file
        """

        struct_name = str(path.split("\\")[-1]).replace(".html", "")

        with open(path, 'r') as reg_file:
            myshit = reg_file.read()

        model = BeautifulSoup(myshit, 'html.parser')
        registers = []

        for table in model.find_all('table', class_="register", recursive=True):
            #print(table)
            my_dict = {}
            for th in table.find_all('th', class_= "address", recursive=True):
                my_dict["address"] = th.text
            for th in table.find_all('th', class_= "register_name", recursive=True):
                my_dict["register_name"] = th.text
            my_dict["fields"] = []
            bitfields = table.find_all('td', class_='fieldname', recursive=True)
            positions = table.find_all('td', class_='register_position', recursive=True)
            defualts = table.find_all('td', class_='reset_value', recursive=True)
            access = table.find_all('td', class_='type', recursive=True)
            descripts = table.find_all('td', class_='description', recursive=True)
            details = table.find_all('td', class_='details', recursive=True)
            for index in range(len(bitfields)):
                my_dict["fields"].append(
                    {
                        "name": bitfields[index].text,
                        "pos": positions[index].text,
                        "default": defualts[index].text,
                        "access": access[index].text,
                        "descSipt": descripts[index].text,
                        "details": details[index].text
                    }
                )
            registers.append(my_dict)

        json_data = {
            struct_name: registers
        }

        with open(path.replace(".html", ".json"), 'w') as dumpfile:
            json.dump(json_data, dumpfile, indent=7)

        return json_data

    @classmethod
    def explainRegisterValue(self, register_name: str, value: int, database: dict) -> dict:
        """
        This method returns a dictionary that shows the value of the bitfields in the register

        Args:
            register_name (str): name of the register to be decoded
            value (int): value of the register
            database (dict): the dictionary containing the information about how to decode the content of the registers

        Returns:
            dict: explaination of the reigster value with a dictionary
        """
        my_register =  self._findRegisterInDatabase(database, register_name)
        return self._decode_register(my_register, value)

    @classmethod
    def _findRegisterInDatabase(self, database: dict, register_name: str) -> dict:
        """
        This method returns the dictionary structure corresponding to the register.

        Args:
            database (dict): database about the content of multiple registers
            register (str): name of the register

        Returns:
            dict: structure belonging to the register
        """

        try:
            database_name = list(database.keys())[0]
            database = database[database_name]
        except AttributeError:
            pass

        for register in database:
            if register["register_name"] == register_name:
                return register

    @classmethod
    def _decode_register(self, reg_dict: dict, value: int) -> dict:
        """
        This method calculates the register value using the dictionary defining the bitfields.
        Implemented for 32 bit registers

        Args:
            reg_dict (dict): dictionary defining the bitfields
            value (int): value of the register

        Returns:
            dict: hash table of bitfield values
        """

        reg_val_inStr = f"{value:032b}"[::-1]
        my_logger.debug(reg_val_inStr)

        explanation  = {}

        for bitfield in reg_dict["fields"]:
            my_logger.debug(bitfield)
            startbit = int(str(bitfield['pos']).split(":")[-1])
            stopbit = int(str(bitfield['pos']).split(":")[0])
            
            my_logger.debug(startbit)
            my_logger.debug(stopbit)
            bitfield_value_str = reg_val_inStr[startbit:stopbit + 1]
            bitfield_value = int(reg_val_inStr[startbit:stopbit + 1], 2)
            

            my_logger.debug(bitfield_value)

            explanation[bitfield["name"]] = bitfield_value


        return explanation

    @classmethod
    def buildRegisterValue(self, database: dict, bitfields: list=None, values: list=None, bitfields_dict: dict=None) -> dict:
        """
        This method returns a list of address - value pairs.
        All value pairs are generated from the bitfield values using the database defining the registers
        If a bitfield of the register is not defined the converter sets it to defult

        Implemented for 32 bit register siye!!!

        Args:
            bitfields (list): bitfields names
            values (list): values of the bitfilds in the same order
            database (dict): register descrition database
            bitfields_dict (dict): the input parameters can be given as dictionary instead of lists by inserting them at this argument
                if this argument is used the bitfields and values arguments will be ignored
                format of this argument:
                    {
                        "bitfield name 1": value as integer
                        "bitfield name 2": value as integer
                        "bitfield name 3": value as integer
                    }

        Returns:
            list: list of register addresses and values
        """
        if bitfields_dict != None:
            bitfields = list(bitfields_dict.keys())
            values = []
            for key in bitfields_dict.keys():
                values.append(bitfields_dict[key])


        my_registers = self._collect_registers(bitfields, database)

        regs_built = {}

        for act_reg in my_registers:
            reg_value = 32 * "0"
            for act_field in act_reg["fields"]:
                bitlength = abs(int(str(act_field["pos"]).split(":")[-1]) - int(str(act_field["pos"]).split(":")[0]))
                startpos = int(str(act_field["pos"]).split(":")[-1])
                startpos = int(str(act_field["pos"]).split(":")[-1])
                my_logger.debug(f"start pos: {startpos}")
                my_logger.debug(f"Bitlength: {bitlength}")
                bitlength = 1 if bitlength == 0 else bitlength
                my_logger.debug(f"Bitlength: {bitlength}")
                mymask = int(bitlength * "1", 2)
                my_logger.debug(f"mask: {mymask:b}")

                if act_field["name"] in bitfields:
                    bf_index = bitfields.index(act_field["name"])
                    my_logger.debug(f"value: 0x{values[bf_index]:x}")
                    my_logger.debug(f"mask: 0x{mymask}")
                    bf_value = values[bf_index] & mymask
                else:
                    base = 16 if "X" in str(act_field["default"]).upper() else 10
                    bf_value = int(act_field["default"], base) & mymask
                             
                my_logger.debug(f"masked value: {bf_value:b}")
                bf_str_value = f"{bf_value:b}"
                reg_value = self.__replace_test_subset_by_position(reg_value, startpos, bf_str_value)
                my_logger.debug(f"New reg value: {reg_value}")
                my_logger.debug(f"New reg value: {int(reg_value[::-1], 2)}")

            regs_built[act_reg["address"]] = int(reg_value[::-1], 2)
        return regs_built

    @classmethod
    def __replace_test_subset_by_position(self, text: str, start: int, new_str: str) -> str:
        """
        This method implements the replacement of the substring in a text by giving the start position and the string.

        Args:
            text (str): original text
            start (int): position where the replacement starts
            new_str (str): text to be given

        Returns:
            str: text with replaced subset
        """
        no_chars = len(new_str)
        return f"{text[0:start]}{new_str}{text[start + no_chars:]}"

    @classmethod
    def _collect_registers(self, bitfields: list, database: dict) -> list:
        """
        This method collects the registers belonging to the bitfields given.

        Args:
            bitfields (list): list of bitfields
            database (dict): description database

        Returns:
            list: list of register description dictionaries
        """
        my_registers = []
        for mybitfield in bitfields:
            act_reg = self._get_register_of_bitfield(mybitfield, database)
            if act_reg not in my_registers:
                my_logger.debug(f"Register appended: {act_reg['register_name']}")
                my_registers.append(act_reg)
        return my_registers

    @classmethod
    def _get_register_of_bitfield(self, bitfield: str, database: dict) -> dict:
        """This method returns the dictionay that describes the register belonging to the bitfield from the database.

        Args:
            bitfield (str): name of the bitfield
            database (dict): register descrition database

        Returns:
            dict: the register belonging to the bitfield
        """
        my_logger.debug(bitfield)
        try:
            database_name = list(database.keys())[0]
            database = database[database_name]
        except AttributeError:
            pass

        for act_reg in database:
            for bf in act_reg["fields"]:
                if bf["name"] == bitfield:
                    my_logger.debug(act_reg["register_name"])
                    return act_reg

    @classmethod
    def ConvertEntireFolder(self, folder_path: str) -> dict:
        """
        This method converts all the file in the given folder and creates one merged file containing all the registers and the merged database is returned as dictionary
        
        Args:
            folder_path (str): path to the folder of the html files

        Returns:
            dict: merged registere description database
        """
        act_cwd = os.getcwd()
        os.chdir(folder_path)

        html_files = glob.glob("*full.html")

        merged_database = []

        for file in html_files:
            my_logger.debug(file)
            db_name = file.replace(".html", "")
            filepath = f"{folder_path}\\{file}"
            act_db = self.convertRegisterFile(filepath)
            act_db = act_db[db_name]
            for register in act_db:
                merged_database.append(register)

        if not self._check_for_duplicate:
            raise KeyError("The register description database contains duplicates...")

        with open(f"{folder_path}\\merged_dataase.json", "w") as merged_file:
            json.dump(merged_database, merged_file, indent= 7)

        os.chdir(act_cwd)

        return merged_database

    @classmethod
    def _check_for_duplicate(self, database: list) -> bool:
        """
        This method checks whether the merged database contains any duplicates in the registers

        Args:
            database (list): description database

        Returns:
            bool: returns true if no duplicate is found
        """
        addresses = []
        for register in database:
            if register["address"] not in addresses:
                addresses.append(register["address"])

        return len(addresses) == len(database)


    @classmethod
    def MSG_dict_2_STRUCT(self, msg_dict: dict) -> register_struct:
        """
        This method converts the dictionary that is usually returned by the conversion functions into an easy to use structure.
        The structure has two attributes:
            - address
            - value

        Args:
            msg_dict (dict): dictionary given by the conversion function

        Returns:
            register_struct: register structure
        """
        my_reg_struct = register_struct(list(msg_dict.keys())[0], msg_dict[list(msg_dict.keys())[0]])
        return my_reg_struct